@@ -2,22 +2,26 @@
 from __future__ import annotations
 
 import asyncio
-import logging
-import pprint
-import uuid
 from collections import defaultdict
+from collections.abc import Callable
 from copy import deepcopy
 from datetime import datetime
+import functools
+import logging
 from operator import itemgetter
+import pprint
 from types import TracebackType
 from typing import Any, cast
+import uuid
 
 from aiohttp import ClientSession, ClientWebSocketResponse, WSMsgType, client_exceptions
 
 from .const import (
+    LOG_LEVEL_MAP,
     MAX_SERVER_SCHEMA_VERSION,
     MIN_SERVER_SCHEMA_VERSION,
     PACKAGE_NAME,
+    LogLevel,
     __version__,
 )
 from .event import Event
@@ -33,8 +37,14 @@ from .exceptions import (
     NotConnected,
 )
 from .model.driver import Driver
+from .model.log_message import LogMessage
 from .model.version import VersionInfo, VersionInfoDataType
 
+try:
+    import orjson as json
+except ImportError:
+    import json
+
 SIZE_PARSE_JSON_EXECUTOR = 8192
 
 # Message IDs
@@ -48,6 +58,9 @@ LISTEN_MESSAGE_IDS = (
     START_LISTENING_MESSAGE_ID,
 )
 
+LOGGER = logging.getLogger(__package__)
+SERVER_LOGGER = logging.getLogger(f"{__package__}.server")
+
 
 class Client:
     """Class to manage the IoT connection."""
@@ -73,10 +86,13 @@ class Client:
             PACKAGE_NAME: __version__,
             **(additional_user_agent_components or {}),
         }
-        self._logger = logging.getLogger(__package__)
         self._loop = asyncio.get_running_loop()
         self._result_futures: dict[str, asyncio.Future] = {}
         self._shutdown_complete_event: asyncio.Event | None = None
+
+        self._server_logger_unsubs: list[Callable[[], None]] = []
+        self._server_logging_enabled: bool = False
+
         self._record_messages = record_messages
         self._recorded_commands: defaultdict[str, dict] = defaultdict(dict)
         self._recorded_events: list[dict] = []
@@ -84,7 +100,10 @@ class Client:
     def __repr__(self) -> str:
         """Return the representation."""
         prefix = "" if self.connected else "not "
-        return f"{type(self).__name__}(ws_server_url={self.ws_server_url!r}, {prefix}connected)"
+        return (
+            f"{type(self).__name__}(ws_server_url={self.ws_server_url!r}, "
+            f"{prefix}connected)"
+        )
 
     @property
     def connected(self) -> bool:
@@ -105,17 +124,18 @@ class Client:
             raise InvalidServerVersion(
                 self.version,
                 require_schema,
-                "Command not available due to incompatible server version. Update the Z-Wave "
-                f"JS Server to a version that supports at least api schema {require_schema}.",
+                "Command not available due to incompatible server version. Update "
+                "the Z-Wave JS Server to a version that supports at least api schema "
+                f"{require_schema}.",
             )
-        future: "asyncio.Future[dict]" = self._loop.create_future()
+        future: asyncio.Future[dict] = self._loop.create_future()
         message_id = message["messageId"] = uuid.uuid4().hex
         self._result_futures[message_id] = future
         await self._send_json_message(message)
         try:
             return await future
         finally:
-            self._result_futures.pop(message_id)
+            self._result_futures.pop(message_id, None)
 
     async def async_send_command_no_wait(
         self, message: dict[str, Any], require_schema: int | None = None
@@ -126,8 +146,9 @@ class Client:
             raise InvalidServerVersion(
                 self.version,
                 require_schema,
-                "Command not available due to incompatible server version. Update the Z-Wave "
-                f"JS Server to a version that supports at least api schema {require_schema}.",
+                "Command not available due to incompatible server version. Update "
+                "the Z-Wave JS Server to a version that supports at least api schema "
+                f"{require_schema}.",
             )
         message["messageId"] = uuid.uuid4().hex
         await self._send_json_message(message)
@@ -137,7 +158,7 @@ class Client:
         if self.driver is not None:
             raise InvalidState("Re-connected with existing driver")
 
-        self._logger.debug("Trying to connect")
+        LOGGER.debug("Trying to connect")
         try:
             self._client = await self.aiohttp_session.ws_connect(
                 self.ws_server_url,
@@ -157,26 +178,26 @@ class Client:
 
         # basic check for server schema version compatibility
         if (
-            self.version.min_schema_version > MIN_SERVER_SCHEMA_VERSION
-            or self.version.max_schema_version < MIN_SERVER_SCHEMA_VERSION
+            version.min_schema_version > MAX_SERVER_SCHEMA_VERSION
+            or version.max_schema_version < MIN_SERVER_SCHEMA_VERSION
         ):
             await self._client.close()
-            assert self.version
+            assert version
             raise InvalidServerVersion(
-                self.version,
+                version,
                 MIN_SERVER_SCHEMA_VERSION,
-                f"Z-Wave JS Server version ({self.version.server_version}) is "
+                f"Z-Wave JS Server version ({version.server_version}) is "
                 "incompatible. Update the Z-Wave JS Server to a version that supports "
                 f"at least api schema {MIN_SERVER_SCHEMA_VERSION}",
             )
         # store the (highest possible) schema version we're going to use/request
-        # this is a bit future proof as we might decide to use a pinned version at some point
-        # for now we just negotiate the highest available schema version and
+        # this is a bit future proof as we might decide to use a pinned version at some
+        # point for now we just negotiate the highest available schema version and
         # guard incompatibility with the MIN_SERVER_SCHEMA_VERSION
-        if self.version.max_schema_version < MAX_SERVER_SCHEMA_VERSION:
-            self.schema_version = self.version.max_schema_version
+        if version.max_schema_version < MAX_SERVER_SCHEMA_VERSION:
+            self.schema_version = version.max_schema_version
 
-        self._logger.info(
+        LOGGER.info(
             "Connected to Home %s (Server %s, Driver %s, Using Schema %s)",
             version.home_id,
             version.server_version,
@@ -203,7 +224,11 @@ class Client:
         if not set_api_msg["success"]:
             # this should not happen, but just in case
             await self._client.close()
-            raise FailedCommand(set_api_msg["messageId"], set_api_msg["errorCode"])
+            raise FailedCommand(
+                set_api_msg["messageId"],
+                set_api_msg["errorCode"],
+                set_api_msg["message"],
+            )
 
     async def listen(self, driver_ready: asyncio.Event) -> None:
         """Start listening to the websocket."""
@@ -227,7 +252,9 @@ class Client:
             # this should not happen, but just in case
             if not log_msg["success"]:
                 await self._client.close()
-                raise FailedCommand(log_msg["messageId"], log_msg["errorCode"])
+                raise FailedCommand(
+                    log_msg["messageId"], log_msg["errorCode"], log_msg["message"]
+                )
 
             # send start_listening command to the server
             # we will receive a full state dump and from now on get events
@@ -239,7 +266,9 @@ class Client:
 
             if not state_msg["success"]:
                 await self._client.close()
-                raise FailedCommand(state_msg["messageId"], state_msg["errorCode"])
+                raise FailedCommand(
+                    state_msg["messageId"], state_msg["errorCode"], state_msg["message"]
+                )
 
             self.driver = cast(
                 Driver,
@@ -254,7 +283,7 @@ class Client:
 
             driver_ready.set()
 
-            self._logger.info(
+            LOGGER.info(
                 "Z-Wave JS initialized. %s nodes", len(self.driver.controller.nodes)
             )
 
@@ -263,7 +292,7 @@ class Client:
             pass
 
         finally:
-            self._logger.debug("Listen completed. Cleaning up")
+            LOGGER.debug("Listen completed. Cleaning up")
 
             for future in self._result_futures.values():
                 future.cancel()
@@ -277,7 +306,7 @@ class Client:
 
     async def disconnect(self) -> None:
         """Disconnect the client."""
-        self._logger.debug("Closing client connection")
+        LOGGER.debug("Closing client connection")
 
         if not self.connected:
             return
@@ -296,6 +325,18 @@ class Client:
         self._shutdown_complete_event = None
         self.driver = None
 
+    async def async_start_listening_logs(self) -> None:
+        """Send command to start listening to log events."""
+        await self.async_send_command(
+            {"command": "start_listening_logs"}, require_schema=31
+        )
+
+    async def async_stop_listening_logs(self) -> None:
+        """Send command to stop listening to log events."""
+        await self.async_send_command(
+            {"command": "stop_listening_logs"}, require_schema=31
+        )
+
     def begin_recording_messages(self) -> None:
         """Begin recording messages for replay later."""
         if self._record_messages:
@@ -319,6 +360,85 @@ class Client:
 
         return list(data)
 
+    @property
+    def server_logging_enabled(self) -> bool:
+        """Return whether server logging is currently enabled."""
+        return self._server_logging_enabled
+
+    async def enable_server_logging(self) -> None:
+        """
+        Enable logging from the server.
+
+        Embeds server logs in the library's logs.
+        """
+        if not self.connected or not self.driver:
+            raise InvalidState(
+                "Can't enable server logging when not connected to server"
+            )
+        if (log_level := self.driver.log_config.level) and (
+            level := LOG_LEVEL_MAP[log_level]
+        ) < LOGGER.level:
+            LOGGER.warning(
+                (
+                    "Server logging is currently more verbose than library logging, "
+                    "setting library log level to %s to match."
+                ),
+                logging.getLevelName(level),
+            )
+            LOGGER.setLevel(level)
+
+        if self._server_logging_enabled:
+            return
+
+        self._server_logging_enabled = True
+
+        def handle_server_logs(event: dict) -> None:
+            """Handle driver `logging` event."""
+            log_msg: LogMessage = event["log_message"]
+            log_level = LogLevel(log_msg.level)
+            SERVER_LOGGER.log(
+                LOG_LEVEL_MAP[log_level],
+                "%s:\n%s",
+                log_msg.timestamp,
+                "\n".join(log_msg.formatted_message),
+            )
+
+        def handle_log_config_updates(event: dict) -> None:
+            """Handle driver `log config updated` events."""
+            if (log_level := event["config"]["level"].lower()) and (
+                level := LOG_LEVEL_MAP[log_level]
+            ) < LOGGER.level:
+                LOGGER.warning(
+                    (
+                        "Server logging is currently more verbose than library "
+                        "logging, setting library log level to %s to match."
+                    ),
+                    logging.getLevelName(level),
+                )
+            LOGGER.setLevel(level)
+
+        self._server_logger_unsubs = [
+            self.driver.on("logging", handle_server_logs),
+            self.driver.on("log config updated", handle_log_config_updates),
+        ]
+
+        await self.async_start_listening_logs()
+
+    async def disable_server_logging(self) -> None:
+        """Disable logging from the server."""
+        if not self.connected or not self.driver:
+            raise InvalidState(
+                "Can't disable server logging when not connected to server"
+            )
+        if not self._server_logging_enabled or not self._server_logger_unsubs:
+            LOGGER.info("Server logging is already disabled")
+            return
+
+        for unsub in self._server_logger_unsubs:
+            unsub()
+        self._server_logger_unsubs.clear()
+        self._server_logging_enabled = False
+
     async def receive_until_closed(self) -> None:
         """Receive messages until client is closed."""
         assert self._client
@@ -344,14 +464,20 @@ class Client:
 
         try:
             if len(msg.data) > SIZE_PARSE_JSON_EXECUTOR:
-                data: dict = await self._loop.run_in_executor(None, msg.json)
+                data: dict = await self._loop.run_in_executor(
+                    None, functools.partial(msg.json, loads=json.loads)
+                )
             else:
-                data = msg.json()
+                data = msg.json(loads=json.loads)
         except ValueError as err:
             raise InvalidMessage("Received invalid JSON.") from err
 
-        if self._logger.isEnabledFor(logging.DEBUG):
-            self._logger.debug("Received message:\n%s\n", pprint.pformat(msg))
+        if LOGGER.isEnabledFor(logging.DEBUG) and not (
+            self.server_logging_enabled
+            and data.get("type") == "event"
+            and data.get("event", {}).get("event") == "logging"
+        ):
+            LOGGER.debug("Received message:\n%s\n", pprint.pformat(msg))
 
         return data
 
@@ -363,8 +489,8 @@ class Client:
         if msg["type"] == "result":
             future = self._result_futures.get(msg["messageId"])
 
-            if future is None:
-                # no listener for this result
+            if future is None or future.cancelled():
+                # no listener for this result or the future was canceled
                 return
 
             if self._record_messages and msg["messageId"] not in LISTEN_MESSAGE_IDS:
@@ -380,7 +506,7 @@ class Client:
                 return
 
             if msg["errorCode"] != "zwave_error":
-                err = FailedCommand(msg["messageId"], msg["errorCode"])
+                err = FailedCommand(msg["messageId"], msg["errorCode"], msg["message"])
             else:
                 err = FailedZWaveCommand(
                     msg["messageId"], msg["zwaveErrorCode"], msg["zwaveErrorMessage"]
@@ -391,14 +517,16 @@ class Client:
 
         if msg["type"] != "event":
             # Can't handle
-            self._logger.debug(
+            LOGGER.debug(
                 "Received message with unknown type '%s': %s",
                 msg["type"],
                 msg,
             )
             return
 
-        if self._record_messages:
+        if self._record_messages and not (
+            self.server_logging_enabled and msg["event"]["event"] == "logging"
+        ):
             self._recorded_events.append(
                 {
                     "record_type": "event",
@@ -419,8 +547,8 @@ class Client:
         if not self.connected:
             raise NotConnected
 
-        if self._logger.isEnabledFor(logging.DEBUG):
-            self._logger.debug("Publishing message:\n%s\n", pprint.pformat(message))
+        if LOGGER.isEnabledFor(logging.DEBUG):
+            LOGGER.debug("Publishing message:\n%s\n", pprint.pformat(message))
 
         assert self._client
         assert "messageId" in message
@@ -437,9 +565,9 @@ class Client:
                 }
             )
 
-        await self._client.send_json(message)
+        await self._client.send_json(message, dumps=json.dumps)
 
-    async def __aenter__(self) -> "Client":
+    async def __aenter__(self) -> Client:
         """Connect to the websocket."""
         await self.connect()
         return self
