@@ -1,19 +1,18 @@
 """Constants for the Z-Wave JS python library."""
 from __future__ import annotations
 
-import logging
 from dataclasses import dataclass, field
 from enum import Enum, IntEnum
-from importlib import metadata
+import logging
 from typing import TypedDict
 
 PACKAGE_NAME = "zwave-js-server-python"
-__version__ = metadata.version(PACKAGE_NAME)
+__version__ = "0.55.0"
 
 # minimal server schema version we can handle
-MIN_SERVER_SCHEMA_VERSION = 32
+MIN_SERVER_SCHEMA_VERSION = 34
 # max server schema version we can handle (and our code is compatible with)
-MAX_SERVER_SCHEMA_VERSION = 32
+MAX_SERVER_SCHEMA_VERSION = 34
 
 VALUE_UNKNOWN = "unknown"
 
@@ -391,7 +390,8 @@ class SetValueStatus(IntEnum):
     FAIL = 2
     # The endpoint specified in the value ID does not exist
     ENDPOINT_NOT_FOUND = 3
-    # The given CC or its API is not implemented (yet) or it has no `setValue` implementation
+    # The given CC or its API is not implemented (yet) or it has no `setValue`
+    # implementation
     NOT_IMPLEMENTED = 4
     # The value to set (or a related value) is invalid
     INVALID_VALUE = 5
@@ -459,7 +459,7 @@ class DateAndTimeDataType(TypedDict, total=False):
 class DateAndTime:
     """Represent a date and time."""
 
-    data: DateAndTimeDataType
+    data: DateAndTimeDataType = field(repr=False)
     hour: int | None = field(init=False)
     minute: int | None = field(init=False)
     weekday: Weekday | None = field(default=None, init=False)
@@ -494,3 +494,13 @@ class ControllerStatus(IntEnum):
     UNRESPONSIVE = 1
     # The controller is unable to transmit
     JAMMED = 2
+
+
+class SupervisionStatus(IntEnum):
+    """Enum for all known supervision statuses."""
+
+    # https://github.com/zwave-js/node-zwave-js/blob/cc_api_options/packages/core/src/consts/Transmission.ts#L304
+    NO_SUPPORT = 0
+    WORKING = 1
+    FAIL = 2
+    SUCCESS = 255
